import {
  useConfig,
  useGraphQueryResult,
  useQuerybuilderGraph,
  useQuerybuilderHash,
  useQuerybuilderSettings,
  useSchemaGraph,
  useSchemaInference,
  useSearchResultQB,
  useUserRole,
} from '@graphpolaris/shared/lib/data-access/store';
import { clearQB, setQuerybuilderGraphology, toQuerybuilderGraphology } from '@graphpolaris/shared/lib/data-access/store/querybuilderSlice';
import React, { useCallback, useEffect, useMemo, useRef, useState } from 'react';
import { useDispatch } from 'react-redux';
import ReactFlow, {
  Background,
  Connection,
  Edge,
  HandleType,
  Node,
  NodeChange,
  NodePositionChange,
  OnConnectStartParams,
  OnEdgesChange,
  ReactFlowInstance,
  ReactFlowProvider,
  isNode,
  useReactFlow,
} from 'reactflow';
import { Dialog, DialogContent } from '../../components/layout/Dialog';
import { addError } from '../../data-access/store/configSlice';
import { toSchemaGraphology } from '../../data-access/store/schemaSlice';
import { LayoutFactory } from '../../graph-layout';
import { AllLogicMap, QueryElementTypes, SchemaReactflowLogicNode, createReactFlowElements, isLogicHandle, toHandleData } from '../model';
import { ConnectionDragLine, ConnectionLine, QueryEntityPill, QueryRelationPill } from '../pills';
import { QueryLogicPill } from '../pills/customFlowPills/logicpill/QueryLogicPill';
import { dragPillStarted, movePillTo } from '../pills/dragging/dragPill';
import styles from './querybuilder.module.scss';
import { QueryBuilderLogicPillsPanel } from './querysidepanel/QueryBuilderLogicPillsPanel';
import { QueryBuilderRelatedNodesPanel } from './querysidepanel/QueryBuilderRelatedNodesPanel';
import { ConnectingNodeDataI } from './utils/connectorDrop';
import { resultSetFocus } from '../../data-access/store/interactionSlice';
import { QueryBuilderDispatcherContext } from './QueryBuilderDispatcher';
import { QueryBuilderNav, QueryBuilderToggleSettings } from './QueryBuilderNav';

export type QueryBuilderProps = {
  onRunQuery?: () => void;
};

/**
 * This is the main querybuilder component. It is responsible for holding all pills and fire off the visual part of the querybuilder panel logic
 */
export const QueryBuilderInner = (props: QueryBuilderProps) => {
  const [toggleSettings, setToggleSettings] = useState<QueryBuilderToggleSettings>();
  const reactFlowWrapper = useRef<HTMLDivElement>(null);
  const queryBuilderSettings = useQuerybuilderSettings();
  const userRole = useUserRole();

  var nodeTypes = useMemo(
    () => ({
      entity: QueryEntityPill,
      relation: QueryRelationPill,
      logic: QueryLogicPill,
    }),
    [],
  );

  var edgeTypes = useMemo(() => ({ connection: ConnectionLine, attribute_connection: ConnectionLine }), []);

  const schemaGraph = useSchemaGraph();
  const schemaInference = useSchemaInference();
  const schema = useMemo(() => toSchemaGraphology(schemaGraph), [schemaGraph]);
  const graph = useQuerybuilderGraph();
  const qbHash = useQuerybuilderHash();
  const config = useConfig();
  const dispatch = useDispatch();
  const isDraggingPill = useRef(false);
  const connectingNodeId = useRef<ConnectingNodeDataI | null>(null);
  const editLogicNode = useRef<SchemaReactflowLogicNode | undefined>(undefined);
  const reactFlow = useReactFlow();
  const isEdgeUpdating = useRef(false);
  const isOnConnect = useRef(false);
  const graphologyGraph = useMemo(() => toQuerybuilderGraphology(graph), [graph, qbHash]);
  const elements = useMemo(() => createReactFlowElements(graphologyGraph), [graph, qbHash]);
  const searchResults = useSearchResultQB();
  const reactFlowInstanceRef = useRef<ReactFlowInstance | null>(null);
  const [allowZoom, setAllowZoom] = useState(true);

  useEffect(() => {
    const searchResultKeys = new Set([...searchResults.nodes.map((node) => node.key), ...searchResults.edges.map((edge) => edge.key)]);

    elements.nodes.forEach((node) => {
      const isNodeInSearchResults = searchResultKeys.has(node?.id);
      if (isNodeInSearchResults && !graphologyGraph.getNodeAttribute(node?.id, 'selected')) {
        graphologyGraph.setNodeAttribute(node?.id, 'selected', true);
      } else if (!isNodeInSearchResults && graphologyGraph.getNodeAttribute(node?.id, 'selected')) {
        graphologyGraph.setNodeAttribute(node?.id, 'selected', false);
      }
    });

    dispatch(setQuerybuilderGraphology(graphologyGraph));
  }, [searchResults]);

  const onInit = (reactFlowInstance: ReactFlowInstance) => {
    setTimeout(() => reactFlow.fitView(), 0);
  };

  /**
   * Clears all nodes in the graph.
   *  TODO: only works if the node is clicked and not moved (maybe use onSelectionChange)
   */
  function onNodesDelete(nodes: Node[]) {
    nodes.forEach((n) => {
      graphologyGraph.dropNode(n.id);
    });

    // Dispatch the new graphology object, so reactflow will get re-rendered
    dispatch(setQuerybuilderGraphology(graphologyGraph));
  }

  /**
   * TODO?
   */
  function onNodesChange(nodes: NodeChange[]) {
    // console.log(nodes);
    nodes.forEach((n) => {
      // console.log(graphologyGraph.findNode((gn) => gn === n?.id));
      if (n.type !== 'position') {
        // updated = true;
        // graphologyGraph.updateAttributes(n.id, n.data);
      } else {
        const node = n as NodePositionChange;
        // Get the node in the elements list to get the previous location
        const pNode = elements.nodes.find((e) => e?.id === node?.id);
        if (!(pNode && isNode(pNode)) || !node?.position) return;
        // This is then used to calculate the delta position
        const dx = node.position.x - pNode.position.x;
        const dy = node.position.y - pNode.position.y;
        // Check if we started dragging, if so, call the drag started usecase
        if (!isDraggingPill.current) {
          dragPillStarted(node.id, graphologyGraph);
          isDraggingPill.current = true;
        }
        // Call the drag usecase
        movePillTo(node.id, graphologyGraph, dx, dy, node.position);
        // Dispatch the new graphology object, so reactflow will get rerendered
        dispatch(setQuerybuilderGraphology(graphologyGraph));
      }
    });
  }

  const onDragOver = (event: React.DragEvent<HTMLDivElement>): void => {
    event.preventDefault();
    event.dataTransfer.dropEffect = 'move';
  };

  const onAddSchemaToQueryBuilder = () => {
    // TODO: incomplete (not being shown) for now due to lack of "OPTIONAL MATCH" support in the backend

    const addedNodes: Record<string, string> = {};
    // loop through all nodes in the schema and add them to the graph
    schemaGraph.nodes.forEach((node) => {
      if (!node.key) return;

      const { id } = graphologyGraph.addPill2Graphology(
        {
          type: QueryElementTypes.Entity,
          // x: position.x - mouse_x,
          // y: position.y - mouse_y,
          x: 0,
          y: 0,
          name: node.key,
          schemaKey: node.key,
          attributes: [],
        },
        schema.getNodeAttribute(node.key, 'attributes'),
        schemaInference,
      );
      if (!id) return;
      addedNodes[node.key] = id;
    });

    // loop all edges in the schema and add them to the graph
    schemaGraph.edges.forEach((edge) => {
      if (!edge.key || !edge.source || !edge.target) return;
      const fromNode = graphologyGraph.getNodeAttributes(addedNodes[edge.source]);
      const toNode = graphologyGraph.getNodeAttributes(addedNodes[edge.target]);
      if (!fromNode || !toNode) return;

      const relation = graphologyGraph.addPill2Graphology(
        {
          type: QueryElementTypes.Relation,
          // x: position.x,
          // y: position.y,
          x: 0,
          y: 0,
          depth: { min: queryBuilderSettings.depth.min, max: queryBuilderSettings.depth.max },
          name: edge.attributes?.collection,
          schemaKey: edge.attributes?.name,
          collection: edge.attributes?.collection,
          attributes: [],
        },
        schema.getEdgeAttribute(edge.key, 'attributes'),
      );

      graphologyGraph.addEdge2Graphology(fromNode, relation, {
        type: 'connection',
        sourceHandleData: toHandleData(edge.source),
        targetHandleData: toHandleData(edge.key),
      });

      graphologyGraph.addEdge2Graphology(relation, toNode, {
        type: 'connection',
        sourceHandleData: toHandleData(edge.key),
        targetHandleData: toHandleData(edge.target),
      });
    });

    dispatch(setQuerybuilderGraphology(graphologyGraph));
  };

  /**
   * The onDrop is called when the user drops an element from the schema onto the QueryBuilder.
   * In the onDrop query elements will be created based on the data stored in the drag event (datastrasfer).
   * @param event Drag event.
   */
  const onDrop = (event: React.DragEvent<HTMLDivElement>): void => {
    if (userRole === 'viewer') {
      event.preventDefault();
      alert('Viewers cannot add nodes and edges to the query builder');

      return;
    }
    event.preventDefault();

    // The dropped element should be a valid reactflow element
    const data: string = event.dataTransfer.getData('application/reactflow');
    if (data.length == 0 || !reactFlow) return;

    const mouse_x = parseFloat(event.dataTransfer.getData('mouse_x'));
    const mouse_y = parseFloat(event.dataTransfer.getData('mouse_y'));

    const dragData = JSON.parse(data);
    const position = reactFlow.screenToFlowPosition({
      x: event.clientX,
      y: event.clientY,
    });

    switch (dragData.type) {
      case QueryElementTypes.Entity:
        // console.log('entity drop', dragData, schema.getNodeAttribute(dragData.name, 'attributes'));
        graphologyGraph.addPill2Graphology(
          {
            type: QueryElementTypes.Entity,
            x: position.x - mouse_x,
            y: position.y - mouse_y,
            name: dragData.name,
            schemaKey: dragData.name,
            attributes: [],
          },
          schema.getNodeAttribute(dragData.name, 'attributes'),
          schemaInference,
        );

        dispatch(setQuerybuilderGraphology(graphologyGraph));
        break;
      // Creates a relation element and will also create the 2 related entities together with the connections
      case QueryElementTypes.Relation:
        const relation = graphologyGraph.addPill2Graphology(
          {
            type: QueryElementTypes.Relation,
            x: position.x,
            y: position.y,
            depth: { min: queryBuilderSettings.depth.min, max: queryBuilderSettings.depth.max },
            name: dragData.collection,
            schemaKey: dragData.label,
            collection: dragData.collection,
            attributes: [],
          },
          schema.getEdgeAttribute(dragData.label, 'attributes'),
        );

        if (config.autoSendQueries) {
          // sendQuery();
        }

        if (queryBuilderSettings.autocompleteRelation == true) {
          let fromNodeID: any = null;
          let toNodeID: any = null;
          schema.nodes().forEach((node: string) => {
            if (node === dragData.from) fromNodeID = node;
            if (node === dragData.to) toNodeID = node;

            if (fromNodeID && toNodeID) return;
          });

          if (fromNodeID && toNodeID) {
            const fromNode = graphologyGraph.addPill2Graphology(
              {
                type: QueryElementTypes.Entity,
                x: position.x - 180,
                y: position.y,
                name: fromNodeID,
                schemaKey: fromNodeID,
                attributes: [],
              },
              schema.getNodeAttribute(fromNodeID, 'attributes'),
            );

            const toNode = graphologyGraph.addPill2Graphology(
              {
                type: QueryElementTypes.Entity,
                x: position.x + 250,
                y: position.y,
                name: toNodeID,
                schemaKey: toNodeID,
                attributes: [],
              },
              schema.getNodeAttribute(toNodeID, 'attributes'),
            );

            graphologyGraph.addEdge2Graphology(fromNode, relation, {
              type: 'connection',
              sourceHandleData: toHandleData(fromNodeID),
              targetHandleData: toHandleData(dragData.collection),
            });

            graphologyGraph.addEdge2Graphology(relation, toNode, {
              type: 'connection',
              sourceHandleData: toHandleData(dragData.collection),
              targetHandleData: toHandleData(toNodeID),
            });
          }
        }

        dispatch(setQuerybuilderGraphology(graphologyGraph));
        break;
      default:
        const logic = AllLogicMap[dragData.value.key];
        const firstLeftLogicInput = logic.input;
        if (!firstLeftLogicInput) return;

        const logicNode = graphologyGraph.addLogicPill2Graphology({
          name: dragData.value.name,
          type: QueryElementTypes.Logic,
          x: position.x,
          y: position.y,
          logic: logic,
          attributes: [],
          inputs: {},
        });

        dispatch(setQuerybuilderGraphology(graphologyGraph));
    }
  };

  const onNodeMouseEnter = (event: React.MouseEvent, node: Node) => {
    // console.log(node);
    // console.log(schema.getNodeAttribute(node.type, 'attributes'));
    setAllowZoom(false);
  };

  const onNodeMouseLeave = (event: React.MouseEvent, node: Node) => {
    // console.log(node);
    setAllowZoom(true);
  };

  const onConnect = useCallback(
    (connection: Connection) => {
      console.log(connection);

      if (!isEdgeUpdating.current) {
        isOnConnect.current = true;
        if (!connection.sourceHandle || !connection.targetHandle) throw new Error('Connection has no source or target');

        if (!graphologyGraph.hasEdge(connection.source, connection.target)) {
          graphologyGraph.addEdge(connection.source, connection.target, {
            type: 'connection',
            sourceHandleData: toHandleData(connection.sourceHandle),
            targetHandleData: toHandleData(connection.targetHandle),
          });
          dispatch(setQuerybuilderGraphology(graphologyGraph));
        }
      }
    },
    [graph],
  );

  const onConnectStart = useCallback(
    (event: React.MouseEvent | React.TouchEvent, params: OnConnectStartParams) => {
      if (userRole === 'viewer') {
        event.preventDefault();
        return;
      }

      event.preventDefault();
      if (!params?.handleId) return;

      let node = graphologyGraph.getNodeAttributes(params.nodeId);
      const handleData = toHandleData(params.handleId);

      connectingNodeId.current = {
        params,
        node,
        position: { x: 0, y: 0 },
        attribute: { handleData: handleData },
      };
    },
    [graph],
  );

  const onConnectEnd = useCallback(
    (event: any) => {
      const targetIsPane = event.target.classList.contains('react-flow__pane');
      if (isOnConnect.current) {
        isOnConnect.current = false;
        return;
      }

      if (targetIsPane && !isEdgeUpdating.current) {
        let clientX: number = 0;
        let clientY: number = 0;

        if ('touches' in event) clientX = event?.touches?.[0]?.clientX;
        else if ('clientX' in event) clientX = event?.clientX;
        if ('touches' in event) clientY = event?.touches?.[0]?.clientY;
        else if ('clientY' in event) clientY = event?.clientY;

        const position = reactFlow.screenToFlowPosition({ x: clientX, y: clientY });
        if (connectingNodeId?.current) connectingNodeId.current.position = position;

        if (!connectingNodeId?.current?.params?.handleId) {
          dispatch(addError('Connection has no source or target handle id'));
          return;
        } else {
          const data = toHandleData(connectingNodeId.current.params.handleId);
          if (isLogicHandle(data.handleType)) setToggleSettings('logic');
          else setToggleSettings('relatedNodes');
        }

        // setToggleSettings('logic');
      }
    },
    [reactFlow.project],
  );

  const onEdgeUpdateStart = useCallback(() => {
    isEdgeUpdating.current = true;
  }, []);

  const onEdgeUpdate = useCallback(
    (oldEdge: Edge, newConnection: Connection) => {
      if (isEdgeUpdating.current) {
        isEdgeUpdating.current = false;

        if (graphologyGraph.hasEdge(oldEdge.id)) {
          graphologyGraph.dropEdge(oldEdge.id);
        }

        if (!newConnection.sourceHandle || !newConnection.targetHandle) throw new Error('Connection has no source or target');
        graphologyGraph.addEdge(newConnection.source, newConnection.target, {
          type: 'connection',
          sourceHandleData: toHandleData(newConnection.sourceHandle),
          targetHandleData: toHandleData(newConnection.targetHandle),
        });
        dispatch(setQuerybuilderGraphology(graphologyGraph));
      }
    },
    [graph],
  );

  const onEdgesChange = (params: OnEdgesChange) => {
    console.log('edges change', params);
  };

  const onEdgeUpdateEnd = useCallback(
    (event: MouseEvent | TouchEvent, edge: Edge, handleType: HandleType) => {
      if (isEdgeUpdating.current) {
        if (graphologyGraph.hasEdge(edge.id)) {
          graphologyGraph.dropEdge(edge.id);
        }
        dispatch(setQuerybuilderGraphology(graphologyGraph));
      }
      isEdgeUpdating.current = false;
    },
    [graph],
  );

  const onNodeContextMenu = (event: React.MouseEvent, node: Node) => {
    if (userRole === 'viewer') {
      event.preventDefault();
      return;
    }
    event.preventDefault();
    graphologyGraph.dropNode(node.id);
    dispatch(setQuerybuilderGraphology(graphologyGraph));
  };

  function applyLayout() {
    if (queryBuilderSettings.layout !== 'manual') {
      const factory = new LayoutFactory();
      factory.createLayout(queryBuilderSettings.layout).layout(graphologyGraph);
      dispatch(setQuerybuilderGraphology(graphologyGraph));
    }
  }

  const fitView = () => {
    if (reactFlowInstanceRef.current) {
      reactFlowInstanceRef.current.fitView();
    }
  };

  useEffect(() => {
    try {
      applyLayout();
    } catch (e) {
      console.error(e);
    }
  }, [queryBuilderSettings]);

  return (
    <QueryBuilderDispatcherContext.Provider
      value={{
        openLogicPillUpdate: (node) => {
          editLogicNode.current = node;
          setToggleSettings('logic');
        },
        openLogicPillCreate: (params, position) => {
          if (!params?.handleId) return;

          let node = graphologyGraph.getNodeAttributes(params.nodeId);
          const handleData = toHandleData(params.handleId);
          connectingNodeId.current = {
            params,
            node,
            position: position || { x: 0, y: 0 },
            attribute: { handleData: handleData },
          };
          setToggleSettings('logic');
        },
      }}
    >
      <div ref={reactFlowWrapper} className="h-full w-full flex flex-col">
        <QueryBuilderNav
          toggleSettings={toggleSettings}
          onFitView={fitView}
          onApplyLayout={() => applyLayout()}
          onRunQuery={() => {
            if (props.onRunQuery) props.onRunQuery();
          }}
          onScreenshot={() => {}}
          onLogic={() => {
            if (toggleSettings === 'logic') setToggleSettings(undefined);
            else setToggleSettings('logic');
          }}
        />

        <Dialog
          open={toggleSettings === 'logic'}
          onOpenChange={(ret) => {
            if (!ret) setToggleSettings(undefined);
          }}
        >
          <DialogContent className="w-fit rounded-none">
            <QueryBuilderLogicPillsPanel
              onClick={(v) => {
                connectingNodeId.current = null;
                editLogicNode.current = undefined;
                setToggleSettings(undefined);
              }}
              reactFlowWrapper={reactFlowWrapper.current}
              title="Available Logic Pills"
              className="max-h-[75vh] min-h-[45vh]"
              onDrag={() => {}}
              connection={connectingNodeId?.current}
              editNode={editLogicNode.current}
            />
          </DialogContent>
        </Dialog>
        <Dialog
          open={toggleSettings === 'relatedNodes'}
          onOpenChange={(ret) => {
            if (!ret) setToggleSettings(undefined);
          }}
        >
          <DialogContent>
            <QueryBuilderRelatedNodesPanel
              onFinished={() => {
                connectingNodeId.current = null;
                setToggleSettings(undefined);
              }}
              reactFlowWrapper={reactFlowWrapper.current}
              title="Related nodes available to add to the query"
              className="min-h-[75vh] max-h-[75vh]"
              connection={connectingNodeId?.current}
            />
          </DialogContent>
        </Dialog>
        <svg height={0}>
          <defs>
            <marker id="arrowIn" markerWidth="9" markerHeight="10" refX={0} refY="5" orient="auto">
              <polygon points="0 0, 9 5, 0 10" />
            </marker>
            <marker id="arrowOut" markerWidth="10" markerHeight="10" refX={0} refY="5" orient="auto">
              <polygon points="0 0, 10 5, 0 10" />
            </marker>
          </defs>
        </svg>
        <ReactFlow
          edges={elements.edges}
          nodes={elements.nodes}
          snapGrid={[10, 10]}
          zoomOnScroll={allowZoom}
          snapToGrid
          nodeTypes={nodeTypes}
          edgeTypes={edgeTypes}
          connectionLineComponent={ConnectionDragLine}
          onMouseDownCapture={() => dispatch(resultSetFocus({ focusType: 'query' }))}
          // connectionMode={ConnectionMode.Loose}
          onInit={(reactFlowInstance) => {
            reactFlowInstanceRef.current = reactFlowInstance;
            onInit(reactFlowInstance);
          }}
          onNodesChange={onNodesChange}
          onDragOver={onDragOver}
          onConnect={onConnect}
          onConnectStart={onConnectStart}
          onConnectEnd={onConnectEnd}
          // onNodeMouseEnter={onNodeMouseEnter}
          // onNodeMouseLeave={onNodeMouseLeave}
          onEdgeUpdate={onEdgeUpdate}
          onEdgeUpdateStart={onEdgeUpdateStart}
          onEdgeUpdateEnd={onEdgeUpdateEnd}
          onDrop={onDrop}
          // onContextMenu={onContextMenu}
          onNodeContextMenu={onNodeContextMenu}
          // onNodesDelete={onNodesDelete}
          // onNodesChange={onNodesChange}
          deleteKeyCode="Backspace"
          className={styles.reactflow}
          proOptions={{ hideAttribution: true }}
        >
          <Background gap={10} size={0.7} />
        </ReactFlow>
      </div>
    </QueryBuilderDispatcherContext.Provider>
  );
};

export const QueryBuilder = (props: QueryBuilderProps) => {
  return (
    <div className="query-panel flex w-full h-full border bg-light">
      <ReactFlowProvider>
        <QueryBuilderInner {...props} />
      </ReactFlowProvider>
    </div>
  );
};

export default QueryBuilder;
