import React, { useEffect } from 'react';
import { MapProps } from '../../mapvis';
import { Button, EntityPill, Input } from '@graphpolaris/shared/lib/components';
import { LayerSettingsComponentType } from '../../mapvis.types';

export function HeatLayerOptions({
  settings,
  graphMetadata,
  updateLayerSettings,
  spatialAttributes,
  updateSpatialAttribute,
}: LayerSettingsComponentType<MapProps>) {
  const layerType = 'heatmap';
  const layerSettings = settings[layerType];

  useEffect(() => {
    if (!layerSettings) {
      const initialSettingsObject = { nodes: {}, edges: {} };

      graphMetadata.nodes.labels.forEach((node) => {
        initialSettingsObject.nodes = {
          ...initialSettingsObject.nodes,
          [node]: {
            size: 10,
            hidden: false,
          },
        };
      });

      graphMetadata.edges.labels.forEach((edge) => {
        initialSettingsObject.edges = {
          ...initialSettingsObject.edges,
          [edge]: {},
        };
      });

      updateLayerSettings({ ...initialSettingsObject });
    }
  }, [graphMetadata, layerType, settings, updateLayerSettings]);

  const handleCollapseToggle = (type: string, itemType: 'nodes' | 'edges') => {
    if (layerSettings) {
      updateLayerSettings({
        [itemType]: {
          ...layerSettings[itemType],
          [type]: {
            ...layerSettings[itemType][type],
            collapsed: !layerSettings[itemType][type]?.collapsed ?? true,
          },
        },
      });
    }
  };

  return (
    layerSettings && (
      <div>
        {graphMetadata.nodes.labels.map((nodeType) => {
          const nodeSettings = layerSettings?.nodes?.[nodeType] || {};

          return (
            <div className="mt-2" key={nodeType}>
              <div className="flex items-center">
                <Button
                  size="2xs"
                  iconComponent={nodeSettings.collapsed ? 'icon-[ic--baseline-arrow-right]' : 'icon-[ic--baseline-arrow-drop-down]'}
                  variant="ghost"
                  onClick={() => handleCollapseToggle(nodeType, 'nodes')}
                />
                <div className="w-3/4 mr-6 cursor-pointer" onClick={() => handleCollapseToggle(nodeType, 'nodes')}>
                  <EntityPill title={nodeType} />
                </div>
              </div>

              {!nodeSettings.collapsed && (
                <div>
                  <Input
                    label="Hidden"
                    type="boolean"
                    value={nodeSettings.hidden ?? false}
                    onChange={(val) => {
                      updateLayerSettings({ nodes: { ...layerSettings.nodes, [nodeType]: { ...nodeSettings, hidden: val } } });
                    }}
                  />
                  <Input
                    inline
                    label="Latitude"
                    type="dropdown"
                    value={settings?.location[nodeType].lat}
                    options={[...spatialAttributes[nodeType]]}
                    disabled={spatialAttributes[nodeType].length < 1}
                    onChange={(val) => updateSpatialAttribute(nodeType, 'lat', val as string)}
                  />
                  <Input
                    inline
                    label="Longitude"
                    type="dropdown"
                    value={settings?.location[nodeType].lon}
                    options={[...spatialAttributes[nodeType]]}
                    disabled={spatialAttributes[nodeType].length < 1}
                    onChange={(val) => updateSpatialAttribute(nodeType, 'lon', val as string)}
                  />
                  <Input
                    label="Size"
                    type="slider"
                    min={0}
                    max={40}
                    step={1}
                    value={nodeSettings.size}
                    onChange={(val) => {
                      updateLayerSettings({ nodes: { ...layerSettings.nodes, [nodeType]: { ...nodeSettings, size: Number(val) } } });
                    }}
                  />
                </div>
              )}
            </div>
          );
        })}
      </div>
    )
  );
}
