import React, { useEffect } from 'react';
import ColorPicker from '@graphpolaris/shared/lib/components/colorComponents/colorPicker';
import { Button, DropdownColorLegend, EntityPill, Icon, Input, RelationPill } from '@graphpolaris/shared/lib/components';
import { MapProps } from '../../mapvis';
import { LayerSettingsComponentType } from '../../mapvis.types';

export function NodeLinkOptions({
  settings,
  graphMetadata,
  updateLayerSettings,
  spatialAttributes,
  updateSpatialAttribute,
}: LayerSettingsComponentType<MapProps>) {
  const layerType = 'nodelink';
  const layerSettings = settings[layerType];

  useEffect(() => {
    if (!layerSettings) {
      const initialSettingsObject = { enableBrushing: false, nodes: {}, edges: {} };

      graphMetadata.nodes.labels.forEach((node) => {
        initialSettingsObject.nodes = {
          ...initialSettingsObject.nodes,
          [node]: {
            colorByAttribute: false,
            colorAttribute: undefined,
            colorAttributeType: undefined,
            hidden: false,
            shape: 'circle',
            color: [Math.floor(Math.random() * 251), Math.floor(Math.random() * 251), 0],
            size: 10,
          },
        };
      });

      graphMetadata.edges.labels.forEach((edge) => {
        initialSettingsObject.edges = {
          ...initialSettingsObject.edges,
          [edge]: {
            hidden: false,
            width: 1,
            sizeAttribute: '',
            fixed: true,
            color: [0, 0, 0],
          },
        };
      });

      updateLayerSettings({ ...initialSettingsObject });
    }
  }, [graphMetadata, settings, updateLayerSettings]);

  const handleCollapseToggle = (type: string, itemType: 'nodes' | 'edges') => {
    if (layerSettings) {
      updateLayerSettings({
        [itemType]: {
          ...layerSettings[itemType],
          [type]: {
            ...layerSettings[itemType][type],
            collapsed: !layerSettings[itemType][type]?.collapsed ?? true,
          },
        },
      });
    }
  };

  return (
    layerSettings && (
      <div>
        {graphMetadata.nodes.labels.map((nodeType) => {
          const nodeSettings = layerSettings?.nodes?.[nodeType] || {};

          return (
            <div className="mt-2" key={nodeType}>
              <div className="flex items-center">
                <Button
                  size="2xs"
                  iconComponent={nodeSettings.collapsed ? 'icon-[ic--baseline-arrow-right]' : 'icon-[ic--baseline-arrow-drop-down]'}
                  variant="ghost"
                  onClick={() => handleCollapseToggle(nodeType, 'nodes')}
                />
                <div className="flex flex-grow mr-2 cursor-pointer" onClick={() => handleCollapseToggle(nodeType, 'nodes')}>
                  <EntityPill title={nodeType} />
                </div>
              </div>

              {!nodeSettings.collapsed && (
                <div>
                  <Input
                    label="Hidden"
                    type="boolean"
                    value={nodeSettings.hidden}
                    onChange={(val) =>
                      updateLayerSettings({ nodes: { ...layerSettings.nodes, [nodeType]: { ...nodeSettings, hidden: val } } })
                    }
                  />
                  <div className="border-t-2 my-2">
                    <span className="font-bold mt-2">Location attributes</span>
                    <Input
                      inline
                      label="Latitude"
                      type="dropdown"
                      value={settings?.location[nodeType]?.lat}
                      options={[...spatialAttributes[nodeType]]}
                      disabled={spatialAttributes[nodeType].length < 1}
                      onChange={(val) => updateSpatialAttribute(nodeType, 'lat', val as string)}
                    />
                    <Input
                      inline
                      label="Longitude"
                      type="dropdown"
                      value={settings?.location[nodeType]?.lon}
                      options={[...spatialAttributes[nodeType]]}
                      disabled={spatialAttributes[nodeType].length < 1}
                      onChange={(val) => updateSpatialAttribute(nodeType, 'lon', val as string)}
                    />
                  </div>

                  <div className="border-t-2 my-2">
                    <div className="flex justify-between">
                      <span className="font-bold">Color</span>
                      {!nodeSettings.colorByAttribute && (
                        <ColorPicker
                          value={nodeSettings.color}
                          updateValue={(val) => {
                            updateLayerSettings({ nodes: { ...layerSettings.nodes, [nodeType]: { ...nodeSettings, color: val } } });
                          }}
                        />
                      )}
                    </div>

                    <div>
                      <Input
                        label="By attribute"
                        type="boolean"
                        value={nodeSettings.colorByAttribute ?? false}
                        onChange={(val) =>
                          updateLayerSettings({ nodes: { ...layerSettings.nodes, [nodeType]: { ...nodeSettings, colorByAttribute: val } } })
                        }
                      />
                      {nodeSettings.colorByAttribute && (
                        <div>
                          <Input
                            inline
                            label="Color based on"
                            type="dropdown"
                            value={nodeSettings.colorAttribute}
                            options={Object.keys(graphMetadata.nodes.types[nodeType]?.attributes)}
                            disabled={!settings.nodes}
                            onChange={(val) =>
                              updateLayerSettings({
                                nodes: {
                                  ...layerSettings.nodes,
                                  [nodeType]: {
                                    ...nodeSettings,
                                    colorAttribute: String(val),
                                    colorAttributeType: graphMetadata.nodes.types[nodeType].attributes[val].dimension,
                                  },
                                },
                              })
                            }
                          />
                          {nodeSettings.colorAttributeType === 'numerical' ? (
                            <div>
                              <DropdownColorLegend
                                value={settings?.colorScale}
                                onChange={(val) =>
                                  updateLayerSettings({
                                    nodes: { ...layerSettings.nodes, [nodeType]: { ...nodeSettings, colorScale: val } },
                                  })
                                }
                              />
                            </div>
                          ) : (
                            <div>Categorical</div>
                          )}
                        </div>
                      )}
                    </div>
                  </div>

                  <div className="border-t-2 my-2">
                    <span className="font-bold mt-2">Shape & Size</span>
                    <Input
                      inline
                      label="Shape"
                      type="dropdown"
                      value={nodeSettings.shape}
                      options={['circle', 'square', 'triangle', 'diamond', 'location', 'star']}
                      disabled={!settings.shape}
                      onChange={(val) =>
                        updateLayerSettings({
                          nodes: { ...layerSettings.nodes, [nodeType]: { ...nodeSettings, shape: String(val) } },
                        })
                      }
                    />
                    <Input
                      label="Size"
                      type="slider"
                      min={0}
                      max={40}
                      step={1}
                      value={nodeSettings.size}
                      onChange={(val) =>
                        updateLayerSettings({
                          nodes: { ...layerSettings.nodes, [nodeType]: { ...nodeSettings, size: Number(val) } },
                        })
                      }
                    />
                  </div>
                </div>
              )}
            </div>
          );
        })}
        {graphMetadata.edges.labels.map((edgeType) => {
          const edgeSettings = layerSettings?.edges?.[edgeType] || {};

          return (
            <div className="mt-2" key={edgeType}>
              <div className="flex items-center">
                <Button
                  size="2xs"
                  iconComponent={edgeSettings.collapsed ? 'icon-[ic--baseline-arrow-right]' : 'icon-[ic--baseline-arrow-drop-down]'}
                  variant="ghost"
                  onClick={() => handleCollapseToggle(edgeType, 'edges')}
                />
                <div className="w-3/4 mr-6 cursor-pointer" onClick={() => handleCollapseToggle(edgeType, 'edges')}>
                  <RelationPill title={edgeType} />
                </div>
              </div>

              {!edgeSettings.collapsed && (
                <div>
                  <Input
                    label="Hidden"
                    type="boolean"
                    value={edgeSettings.hidden ?? false}
                    onChange={(val) => {
                      updateLayerSettings({ edges: { ...settings.edges, [edgeType]: { ...edgeSettings, hidden: val } } });
                    }}
                  />

                  <div className="flex justify-between">
                    <span className="font-bold">Color</span>
                    <ColorPicker
                      value={edgeSettings.color}
                      updateValue={(val) =>
                        updateLayerSettings({ edges: { ...settings.edges, [edgeType]: { ...edgeSettings, color: val } } })
                      }
                    />
                  </div>

                  <Input
                    label="Enable brushing"
                    type="boolean"
                    value={settings.enableBrushing}
                    onChange={(val) => {
                      updateLayerSettings({ enableBrushing: val as boolean });
                    }}
                  />

                  <div>
                    <div className="flex items-center gap-1">
                      <Icon component="icon-[ic--baseline-subdirectory-arrow-right]" size={16} color="text-secondary-300" />
                      <span>Width</span>
                    </div>
                    <Input
                      label="Fixed"
                      type="boolean"
                      value={edgeSettings.fixed}
                      onChange={(val) => updateLayerSettings({ edges: { ...settings.edges, [edgeType]: { ...edgeSettings, fixed: val } } })}
                    />
                    {!edgeSettings.fixed ? (
                      <div>
                        <Input
                          label="Based on"
                          type="dropdown"
                          size="xs"
                          options={
                            graphMetadata.edges.types[edgeType]?.attributes
                              ? Object.keys(graphMetadata.edges.types[edgeType].attributes).filter(
                                  (key) => graphMetadata.edges.types[edgeType].attributes[key].dimension === 'numerical',
                                )
                              : []
                          }
                          value={edgeSettings.sizeAttribute}
                          onChange={(val) =>
                            updateLayerSettings({
                              edges: { ...settings.edges, [edgeType]: { ...edgeSettings, sizeAttribute: String(val) } },
                            })
                          }
                        />
                        <div className="flex">
                          <Input
                            type="number"
                            label="min"
                            size="xs"
                            value={edgeSettings.min}
                            onChange={(val) =>
                              updateLayerSettings({ edges: { ...settings.edges, [edgeType]: { ...edgeSettings, min: val } } })
                            }
                          />
                          <Input
                            type="number"
                            label="max"
                            size="xs"
                            value={edgeSettings.max}
                            onChange={(val) =>
                              updateLayerSettings({ edges: { ...settings.edges, [edgeType]: { ...edgeSettings, max: val } } })
                            }
                          />
                        </div>
                      </div>
                    ) : (
                      <div>
                        <Input
                          type="slider"
                          label="Width"
                          min={0}
                          max={10}
                          step={0.2}
                          value={edgeSettings.width}
                          onChange={(val) =>
                            updateLayerSettings({ edges: { ...settings.edges, [edgeType]: { ...edgeSettings, width: Number(val) } } })
                          }
                        />
                      </div>
                    )}
                  </div>
                </div>
              )}
            </div>
          );
        })}
      </div>
    )
  );
}
